import typer
from pathlib import Path

from console_tea.commands.config import app as config_app

from traktor_server.config import config


app = typer.Typer(name="traktor-server", help="Traktor server.")


# Add tea subcommands
app.add_typer(config_app)


@app.callback()
def callback(
    config_path: Path = typer.Option(
        default=None,
        help="Path to the configuration.",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=True,
        readable=True,
        resolve_path=True,
    ),
):
    if config_path is not None:
        config.config_path = config_path

    config.load()


@app.command()
def runserver():
    """Run development server."""
    from django.core.management import execute_from_command_line

    execute_from_command_line(
        ["traktor-server", "runserver", config.server_url]
    )


@app.command()
def gunicorn():
    """Run gunicorn server."""
    from gunicorn.app import base
    from traktor_server import wsgi

    class TraktorApplication(base.BaseApplication):
        def __init__(self, application, **kwargs):
            self.options = kwargs
            self.application = application
            super().__init__()

        def load_config(self):
            config_dict = {
                key: value
                for key, value in self.options.items()
                if key in self.cfg.settings and value is not None
            }
            for key, value in config_dict.items():
                self.cfg.set(key.lower(), value)

        def init(self, parser, opts, args):
            pass

        def load(self):
            return self.application

    options = {
        "bind": f"0.0.0.0:{config.server_port}",
        "workers": config.server_workers,
    }
    TraktorApplication(wsgi.application, **options).run()


@app.command(hidden=True)
def shell():
    """Run IPython shell with loaded configuration and models."""
    try:
        from IPython import embed
        from traktor.config import config
        from traktor.engine import engine
        from traktor.models import Project, Task, Entry

        embed(
            user_ns={
                "config": config,
                "engine": engine,
                "Project": Project,
                "Task": Task,
                "Entry": Entry,
            },
            colors="neutral",
        )
    except ImportError:
        typer.secho("IPython is not installed", color=typer.colors.RED)



from traktor.config import ConfigField, Config as TraktorConfig


class Config(TraktorConfig):
    ENTRIES = {
        **TraktorConfig.ENTRIES,
        "server_host": ConfigField(section="server", option="host"),
        "server_port": ConfigField(section="server", option="port", type=int),
        "server_workers": ConfigField(
            section="server", option="workers", type=int
        ),
    }

    def __init__(self):
        self.server_host = "127.0.0.1"
        self.server_port = 8080
        self.server_workers = 2

        super().__init__()

    @property
    def server_url(self):
        return f"{self.server_host}:{self.server_port}"


config = Config()



from django_tea.consts import COLOR_RE
from rest_framework.exceptions import ValidationError


def color_validator(value):
    if COLOR_RE.match(value) is None:
        raise ValidationError(f"Invalid color string: {value}")
    return value



"""traktor URL Configuration."""

from django.contrib import admin
from django.urls import path, include
from django.contrib.staticfiles.urls import staticfiles_urlpatterns

urlpatterns = [
    # admin
    path("admin/", admin.site.urls),
    # auth
    path("accounts/", include("django.contrib.auth.urls")),
    # api v0
    path("api/v0/", include("traktor_server.views.api.v0.urls")),
] + staticfiles_urlpatterns()



import os
import sys
from pathlib import Path


def main():
    # Setup environment

    # Setup python path
    module_dir = str(Path(__file__).parents[1].absolute())
    if module_dir not in sys.path:
        sys.path.insert(0, module_dir)
    python_path = os.environ.get("PYTHONPATH", "").strip()
    if python_path == "":
        os.environ.setdefault("PYTHONPATH", module_dir)
    else:
        if module_dir not in python_path.split(":"):
            os.environ.setdefault(
                "PYTHONPATH", f"{module_dir}:{python_path.lstrip(':')}"
            )

    # Setup django setting module
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "traktor_server.settings")

    # Setup django
    import django

    django.setup()

    args = sys.argv[:]
    # Check if it's a django manage command and run it
    if (
        len(args) >= 2
        and args[0].endswith("traktor-server")
        and args[1] == "manage"
    ):
        from django.core.management import execute_from_command_line

        execute_from_command_line(args[:1] + args[2:])
    # If not run the application
    else:
        from traktor_server.commands import app

        app()


if __name__ == "__main__":
    main()



from django.urls import path, include

from traktor_server.views.api.v0 import api
from traktor_server.views.api.v0 import auth


urlpatterns = [
    # Auth
    path("auth/token/", auth.ObtainToken.as_view()),
    path("auth/token/refresh/", auth.RefreshToken.as_view()),
    path("auth/", include("rest_framework.urls")),
    # Project
    path("projects/", api.ProjectListCreate.as_view()),
    path("projects/<slug:project_id>/", api.ProjectGetUpdateDelete.as_view()),
    # Task
    path("projects/<slug:project_id>/tasks/", api.TaskListCreate.as_view()),
    path(
        "projects/<slug:project_id>/tasks/<slug:task_id>/",
        api.TaskGetUpdateDelete.as_view(),
    ),
    # Timer
    path("timer/start/<slug:project_id>/", api.timer_default_start),
    path("timer/start/<slug:project_id>/<slug:task_id>/", api.timer_start),
    path("timer/stop/", api.timer_stop),
    path("timer/status/", api.timer_status),
    path("timer/today/", api.timer_today),
    path("timer/report/", api.timer_report),
]



"""
Django settings for traktor project.

Generated by 'django-admin startproject' using Django 3.1.

For more information on this file, see
https://docs.djangoproject.com/en/3.1/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/3.1/ref/settings/
"""

from pathlib import Path
from datetime import timedelta

from traktor_server import log
from traktor_server.config import config


# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve(strict=True).parent.parent

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/3.1/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = config.secret_key

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = config.debug

ALLOWED_HOSTS = [config.server_host]


# Application definition

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "rest_framework",
    "django_tea",
    "traktor",
    "traktor_server",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "traktor_server.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "traktor_server.wsgi.application"


# Database
# https://docs.djangoproject.com/en/3.1/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": f"django.db.backends.{config.db_engine}",
        "NAME": config.db_name,
        "USER": config.db_user,
        "PASSWORD": config.db_password,
        "HOST": config.db_host,
        "PORT": config.db_port,
    }
}


AUTH_USER_MODEL = "traktor.User"

# Password validation
# https://docs.djangoproject.com/en/3.1/ref/settings/#auth-password-validators
AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation"
        ".UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation"
        ".MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation"
        ".CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation"
        ".NumericPasswordValidator",
    },
]


# Internationalization
# https://docs.djangoproject.com/en/3.1/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = config.timezone.zone

USE_I18N = True

USE_L10N = True

USE_TZ = True

LOGGING = log.create_logging_configuration(
    log_dir=config.log_dir, hosted=not config.debug
)


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.1/howto/static-files/

STATIC_URL = "/static/"

# Rest Framework
REST_FRAMEWORK = {
    # Use Django's standard `django.contrib.auth` permissions,
    # or allow read-only access for unauthenticated users.
    "DEFAULT_PERMISSION_CLASSES": [
        "rest_framework.permissions.IsAuthenticated"
    ],
    "DEFAULT_PARSER_CLASSES": ["rest_framework.parsers.JSONParser"],
    "DEFAULT_RENDERER_CLASSES": [
        "rest_framework.renderers.JSONRenderer",
        "rest_framework.renderers.BrowsableAPIRenderer",
    ],
    "DEFAULT_AUTHENTICATION_CLASSES": [
        "rest_framework_jwt.authentication.JSONWebTokenAuthentication",
        "rest_framework.authentication.SessionAuthentication",
    ],
    "EXCEPTION_HANDLER": (
        "traktor_server.exception_handler.traktor_exception_handler"
    ),
}

JWT_AUTH = {
    # "JWT_SECRET_KEY": SECRET_KEY,
    # "JWT_ALGORITHM": "HS256",
    # "JWT_VERIFY": True,
    # "JWT_VERIFY_EXPIRATION": True,
    "JWT_EXPIRATION_DELTA": timedelta(days=7),
    # "JWT_ALLOW_REFRESH": True,
    "JWT_REFRESH_EXPIRATION_DELTA": timedelta(days=30),
}



from rest_framework import serializers

from traktor.engine import engine
from traktor.models import Project, Task, Entry
from traktor_server.serializers.validators import color_validator

# Colored


class ColoredCreateSerializer(serializers.Serializer):
    color = serializers.CharField(max_length=7, validators=[color_validator])


class ColoredUpdateSerializer(serializers.Serializer):
    color = serializers.CharField(
        max_length=7,
        validators=[color_validator],
        required=False,
        allow_null=True,
    )

    def update(self, instance, validated_data):
        color = validated_data.get("color", None)
        if color is not None:
            instance.color = color


# Project


class ProjectSerializer(serializers.ModelSerializer):
    id = serializers.SerializerMethodField()

    def get_id(self, obj):
        return obj.slug

    class Meta:
        model = Project
        fields = ["id", "name", "color", "created_on", "updated_on"]


class ProjectCreateSerializer(ColoredCreateSerializer):
    name = serializers.CharField(max_length=255)

    def create(self, validated_data):
        return engine.project_create(**validated_data)


class ProjectUpdateSerializer(ColoredUpdateSerializer):
    name = serializers.CharField(
        max_length=255, required=False, allow_null=True
    )


# Task


class TaskSerializer(serializers.ModelSerializer):
    project = serializers.SerializerMethodField()
    id = serializers.SerializerMethodField()

    def get_project(self, obj):
        return obj.project.slug

    def get_id(self, obj):
        return obj.slug

    class Meta:
        model = Task
        fields = [
            "project",
            "id",
            "name",
            "color",
            "default",
            "created_on",
            "updated_on",
        ]


class TaskCreateSerializer(ColoredCreateSerializer):
    name = serializers.CharField(max_length=255)
    default = serializers.BooleanField(default=False, required=False)


class TaskUpdateSerializer(ColoredUpdateSerializer):
    name = serializers.CharField(
        max_length=255, required=False, allow_null=True
    )
    default = serializers.BooleanField(
        default=False, required=False, allow_null=True
    )


# Entry


class EntrySerializer(serializers.ModelSerializer):
    project = serializers.SerializerMethodField()
    task = serializers.SerializerMethodField()
    running_time = serializers.SerializerMethodField()

    def get_project(self, obj) -> str:
        return obj.project.slug

    def get_task(self, obj) -> str:
        return obj.task.slug

    def get_running_time(self, obj) -> str:
        return obj.running_time

    class Meta:
        model = Entry
        fields = [
            "project",
            "task",
            "description",
            "notes",
            "start_time",
            "end_time",
            "duration",
            "running_time",
            "created_on",
            "updated_on",
        ]


# Report


class ReportSerializer(serializers.Serializer):
    project = serializers.CharField()
    task = serializers.CharField()
    duration = serializers.IntegerField()
    running_time = serializers.CharField()



import os


def create_logging_configuration(log_dir: str, hosted: bool = False):
    """Create logging configuration.

    Args:
        log_dir: Path to the log directory
        hosted: Is this hosted or development environment.
    """
    # Create if it doesn't exist
    os.makedirs(log_dir, exist_ok=True)

    max_bytes = 1024 * 1024 * 1024  # max log size: 1 GB

    return {
        "version": 1,
        "disable_existing_loggers": True,
        "formatters": {
            "verbose": {
                "format": (
                    "{asctime} {levelname} {name} {message} "
                    "fileName={filename} func={funcName} lineNum={lineno} "
                    "threadName={threadName} process={process}"
                ),
                "style": "{",
            },
            "simple": {
                "format": "{asctime} {levelname} {message} {module}",
                "style": "{",
            },
        },
        "handlers": {
            "traktor": {
                "level": "DEBUG",
                "class": "logging.handlers.RotatingFileHandler",
                "filename": os.path.join(log_dir, "traktor.log"),
                "maxBytes": max_bytes,
                "backupCount": 5,
                "formatter": "verbose",
            },
            "traktor-error": {
                "level": "ERROR",
                "class": "logging.handlers.RotatingFileHandler",
                "filename": os.path.join(log_dir, "traktor-error.log"),
                "maxBytes": max_bytes,
                "backupCount": 5,
                "formatter": "verbose",
            },
            "traktor-console": {
                "level": "INFO",
                "class": "logging.StreamHandler",
                "formatter": "simple",
            },
        },
        "loggers": {
            "": {
                "level": "INFO",
                "handlers": ["traktor", "traktor-error"]
                + ([] if hosted else ["traktor-console"]),
            },
            "traktor": {
                "level": "INFO",
                "handlers": ["traktor", "traktor-error"]
                + ([] if hosted else ["traktor-console"]),
                "propagate": False,
            },
            # 3rd Party Software Logs
            "django": {
                "level": "INFO",
                "handlers": ["traktor", "traktor-error"]
                + ([] if hosted else ["traktor-console"]),
                "propagate": True,
            },
            "django.request": {
                "level": "ERROR",
                "handlers": ["traktor", "traktor-error"]
                + ([] if hosted else ["traktor-console"]),
                "propagate": False,
            },
            "django.server": {
                "level": "INFO",
                "handlers": ["traktor", "traktor-error"]
                + ([] if hosted else ["traktor-console"]),
                "propagate": False,
            },
            "django.db.backends": {
                "level": "ERROR",
                "handlers": ["traktor", "traktor-error"]
                + ([] if hosted else ["traktor-console"]),
                "propagate": False,
            },
        },
    }



from rest_framework.exceptions import ValidationError
from rest_framework_jwt.views import ObtainJSONWebToken, RefreshJSONWebToken


class ObtainToken(ObtainJSONWebToken):
    def post(self, request, *args, **kwargs):
        response = super().post(request, *args, **kwargs)
        if response.status_code == 400:
            raise ValidationError(detail=response.data)
        return response


class RefreshToken(RefreshJSONWebToken):
    def post(self, request, *args, **kwargs):
        response = super().post(request, *args, **kwargs)
        if response.status_code == 400:
            raise ValidationError(detail=response.data)
        return response



import functools
from typing import Optional, Type

from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.serializers import Serializer
from rest_framework.decorators import api_view, APIView


from traktor_server.serializers import (
    ProjectSerializer,
    ProjectCreateSerializer,
    ProjectUpdateSerializer,
    TaskSerializer,
    TaskCreateSerializer,
    TaskUpdateSerializer,
    EntrySerializer,
    ReportSerializer,
)
from traktor.engine import engine


# Project


def handler(serializer: Optional[Type[Serializer]] = None, many=False):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            if serializer is None:
                return Response(result)
            else:
                return Response(serializer(result, many=many).data)

        return wrapper

    return decorator


class ProjectListCreate(APIView):
    @handler(ProjectSerializer, many=True)
    def get(self, request: Request):
        """List all projects."""
        return engine.project_list()

    @handler(ProjectSerializer)
    def post(self, request: Request):
        """Create a project."""
        serializer = ProjectCreateSerializer(data=request.data)
        if serializer.is_valid(raise_exception=True):
            return engine.project_create(**serializer.data)


class ProjectGetUpdateDelete(APIView):
    @handler(ProjectSerializer)
    def get(self, request: Request, project_id: str):
        """Get a project."""
        return engine.project_get(project_id=project_id)

    @handler(ProjectSerializer)
    def patch(self, request: Request, project_id: str):
        """Update a project."""
        project = engine.project_get(project_id=project_id)
        serializer = ProjectUpdateSerializer(data=request.data)
        if serializer.is_valid(raise_exception=True):
            return engine.project_update(project.slug, **serializer.data)

    def delete(self, request: Request, project_id: str):
        """Delete a project."""
        engine.project_delete(project_id=project_id)
        return Response(status=204, data={"detail": "OK"})


# Task


class TaskListCreate(APIView):
    @handler(TaskSerializer, many=True)
    def get(self, request: Request, project_id: str):
        """List all tasks in project.."""
        return engine.task_list(project_id=project_id)

    @handler(TaskSerializer)
    def post(self, request: Request, project_id: str):
        """Create a task.."""
        serializer = TaskCreateSerializer(data=request.data)
        if serializer.is_valid(raise_exception=True):
            return engine.task_create(project_id=project_id, **serializer.data)


class TaskGetUpdateDelete(APIView):
    @handler(TaskSerializer)
    def get(self, request: Request, project_id: str, task_id: str):
        """Get task details."""
        return engine.task_get(project_id=project_id, task_id=task_id)

    @handler(TaskSerializer)
    def patch(self, request: Request, project_id: str, task_id: str):
        """Update a task."""
        serializer = TaskUpdateSerializer(data=request.data)
        if serializer.is_valid(raise_exception=True):
            return engine.task_update(
                project_id=project_id, task_id=task_id, **serializer.data
            )

    def delete(self, request: Request, project_id: str, task_id: str):
        """Delete a task."""
        engine.task_delete(project_id=project_id, task_id=task_id)
        return Response(status=204, data={"detail": "OK"})


# Timer


@api_view(["POST"])
@handler(EntrySerializer)
def timer_default_start(requests: Request, project_id: str):
    return engine.timer_start(project_id=project_id)


@api_view(["POST"])
@handler(EntrySerializer)
def timer_start(request: Request, project_id: str, task_id: str):
    return engine.timer_start(project_id=project_id, task_id=task_id)


@api_view(["POST"])
@handler(EntrySerializer)
def timer_stop(request: Request):
    return engine.timer_stop()


@api_view(["GET"])
@handler(EntrySerializer)
def timer_status(request: Request):
    return engine.timer_status()


@api_view(["GET"])
@handler(ReportSerializer, many=True)
def timer_today(request: Request):
    return engine.timer_today()


@api_view(["GET"])
@handler(ReportSerializer, many=True)
def timer_report(request: Request):
    days = request.query_params.get("days", 0)
    return engine.timer_report(days=days)



__all__ = [
    "ProjectSerializer",
    "ProjectCreateSerializer",
    "ProjectUpdateSerializer",
    "TaskSerializer",
    "TaskCreateSerializer",
    "TaskUpdateSerializer",
    "EntrySerializer",
    "ReportSerializer",
]

from traktor_server.serializers.serializers import (
    ProjectSerializer,
    ProjectCreateSerializer,
    ProjectUpdateSerializer,
    TaskSerializer,
    TaskCreateSerializer,
    TaskUpdateSerializer,
    EntrySerializer,
    ReportSerializer,
)
